# io.s
#

################################################################################
##                                 COSTANTI                                   ##
################################################################################

# Tipi delle interruzioni delle chiamate di sistema usate
# Devono coincidere con quelli usati in sistema.s e utente.s
#
.set tipo_t, 0x31
.set tipo_si, 0x34
.set tipo_w, 0x35
.set tipo_s, 0x36

# Tipi delle interruzioni usate per interfacciarsi al modulo
#  sistema
# Devono coincidere con quelli usati in sistema.s
#
.set tipo_d, 0x39
.set tipo_ae, 0x40
.set tipo_nwfi, 0x41
.set tipo_va, 0x42
.set tipo_p, 0x43
.set tipo_fg, 0x44
.set tipo_r, 0x45
.set tipo_tras, 0x46
.set tipo_cr, 0x50
.set tipo_cw, 0x51
.set tipo_cs, 0x52
.set tipo_cl, 0x53
.set tipo_cu, 0x54
.set tipo_ci, 0x55
.set tipo_pk, 0x56

# Tipi di interruzioni usate per lo swap in/out di un'area
#  di memoria su dsco
#
.set io_tipo_swp, 0x5F

# Tipi delle interruzioni usate dalle primitive di IO
# Devono coincidere con quelli usati in utente.s
#
.set io_tipo_rsen, 0x60
.set io_tipo_rseln, 0x61
.set io_tipo_wsen, 0x62
.set io_tipo_wse0, 0x63
.set io_tipo_tr, 0x64
.set io_tipo_tw, 0x65
.set io_tipo_geom, 0x66
.set io_tipo_rhdn, 0x67
.set io_tipo_whdn, 0x68
.set io_tipo_rfdn, 0x69
.set io_tipo_wfdn, 0x6A

#interruzioni aggiunte per scheda audio
.set io_tipo_dp, 0x6B
.set io_tipo_gams, 0x6C
.set io_tipo_sams, 0x6D
.set io_tipo_fdma, 0x6E
.set io_tipo_isnd, 0x6F
.set io_tipo_wsb, 0x80
.set io_tipo_rsb, 0x81
.set io_tipo_sms, 0x82
.set io_tipo_gms, 0x83
.set io_tipo_spon, 0x84
.set io_tipo_spof, 0x85
.set io_tipo_dpl, 0x86
.set io_tipo_drd, 0x87
.set io_tipo_gss, 0x88
.set io_tipo_rmd, 0x89
.set io_tipo_wmd, 0x8A
.set io_tipo_rsmd, 0x8B
.set io_tipo_stmd, 0x8C

################################################################################
##                  MACRO PER LA MANIPOLAZIONE DEI PARAMETRI                  ##
################################################################################

# Copia dei parametri di una chiamata di sistema dalla pila utente
#  alla pila sistema
# Replicato in sistema.s
#
.macro copia_param n_long offset
        movl $\offset, %ecx
        movl 4(%esp, %ecx, 4), %eax     # cs in eax
        testl $3, %eax			# verifica del livello di privilegio
					#  del chiamante
        jz 1f                           # copia da pila sistema

        movl 12(%esp, %ecx, 4), %eax    # vecchio esp (della pila utente)
					#  in eax
        leal 4(%eax), %esi		# indirizzo del primo parametro in
					#  esi
        jmp 2f
1:
        leal 16(%esp, %ecx, 4), %esi	# indirizzo del primo parametro in esi
2:
        movl $\n_long, %eax		# creazione in pila dello spazio per
        shll $2, %eax			#  la copia dei parametri
        subl %eax, %esp
        leal (%esp), %edi		# indirizzo della destinazione del
					#  primo parametro in edi

        movl $\n_long, %ecx
        cld
        rep
           movsl			# copia dei parametri
.endm

# Salvataggio dei registri in pila
# Replicato in sistema.s
#
.macro salva_registri
	pushl %eax
	pushl %ecx
	pushl %edx
	pushl %ebx
	pushl %esi
	pushl %edi
	pushl %ebp
.endm

# Caricamento dei registri dalla pila (duale rispetto a salva_registri)
# Replicato in sistema.s
#
.macro carica_registri
	popl %ebp
	popl %edi
	popl %esi
	popl %ebx
	popl %edx
	popl %ecx
	popl %eax
.endm

################################################################################
##                             SEZIONE DATI                                   ##
################################################################################

	.data

# Descrittori delle interfacce seriali
#
	.global _com		# non comi, como
_com:	.long	0x03f8		# com[0].indreg.iRBR
	.long	0x03f8		# com[0].indreg.iTHR
	.long	0x03fd		# com[0].indreg.iLSR
	.long	0x03f9		# com[0].indreg.iIER
	.long	0x03fa		# com[0].indreg.iIIR
	.long	0		# com[0].mutex
	.long	0		# com[0].sincr
	.long	0		# com[0].cont
	.long	0		# com[0].punt
	.long	0		# com[0].funzione
	.long	0		# com[0].stato
	.long	0x02f8		# com[1].indreg.iRBR
	.long	0x02f8		# com[1].indreg.iTHR
	.long	0x03fd		# com[1].indreg.iLSR
	.long	0x02f9		# com[1].indreg.iIER
	.long	0x02fa		# com[1].indreg.iIIR
	.long	0		# com[1].mutex
	.long	0		# com[1].sincr
	.long	0		# com[1].cont
	.long	0		# com[1].punt
	.long	0		# com[1].funzione
	.long	0		# com[1].stato

# Descrittori dei Controllori DMA (Registri comuni)
#

	.global _cdma
_cdma:	.long	0x0008		#cdma[0].iCMD_STS
	.long	0x000B		#cdma[0].iMODER
	.long	0x0009		#cdma[0].iREQREG
	.long	0x000F		#cdma[0].iMASK_REG
	.long	0x000A		#cdma[0].iMASK
	.long	0x000C		#cdma[0].CLEAR_FF
	.long	0x000D		#cdma[0].MASTERRS
	.long	0x000E		#cdma[0].MASKCLR
	.long	0x00D0		#cdma[1].iCMD_STS
	.long	0x00D6		#cdma[1].iMODER
	.long	0x00D2		#cdma[1].iREQREG
	.long	0x00DE		#cdma[1].iMASK_REG
	.long	0x00D4		#cdma[1].iMASK
	.long	0x00D8		#cdma[1].CLEAR_FF
	.long	0x00DA		#cdma[1].MASTERRS
	.long	0x00DC		#cdma[1].MASKCLR

# Descrittore del Floppy Disk
#

	.global _fd
_fd:	.long	0x03F2		#fd.indreg.iDOR
	.long	0x03F4		#fd.indreg.iM_STS_D_RATE
	.long	0x03F5		#fd.indreg.iDATA
	.long	0x03F7		#fd.indreg.iDIR_CCR
	.long	0x0081		#fd.dmareg.iPAGE
	.long	0x0004		#fd.dmareg.iMAR
	.long	0x0005		#fd.dmareg.iTCR
	.long	0x000C		#FLIP FLOP	
	.long	0		#fd.drive[0].{motore,presente,errore} (byte!)
	.long	0		#fd.drive[0].traccia
	.long	0		#fd.drive[1].{motore,presente,errore} (byte!)
	.long	0		#fd.drive[1].traccia
	.long	0		#fd.curr_cmd
	.long	0		#fd.curr_drv
	.long	0		#fd.mutex
	.long	0		#fd.sincr
	.long	2		#fd_dma

	.global _sb_16
_sb_16:	.long	0x0220		#
	.long	0x0221		#
	.long	0x0222		#
	.long	0x0223		#
	.long	0x0224		#
	.long	0x0225		#
	.long	0x0226		#
	.long	0x0227		#
	.long	0x0228		#
	.long	0x0229		#
	.long	0x022A		#
	.long	0x022B		#
	.long	0x022C		#
	.long	0x022D		#
	.long	0x022E		#
	.long	0x022F		#
	.long	0x0230		#
	.long	0x0231		#
	.long	0x0232		#
	.long	0x0233		#
	.long	0x0310		
	.long	0x0311		
	.long	0x0083		#sb_16.dmareg.iPAGE
	.long	0x0002		#sb_16.dmareg.iMAR
	.long	0x0003		#sb_16.dmareg.iTCR
	.long	0x000C		#FLIP FLOP
	.long	0		#sb_16.mutex
	.long	0		#sb_16.sincr
	.long	0		#sb_16.ModeBits
	.long	0		#sb_16.ModeStereoMono	
	.long	0		#sb_16.SoundRate
	.long	0		#sb_16.TransferLenght
	.long	0		#dma
	.long	0		#irq
	.long	0		#dma16
	.long	0		#transfermode
	.long	0		#done
	.long	0		#soundsize
	.long	0		#place
	.long	0		#midi
	
# Descrittore Hard Disk
#

	.global _hd
_hd:	.long	0x01f7		#hd[0].indreg.iCMD_iSTS
	.long	0x01f0		#hd[0].indreg.iDATA
	.long	0x01f1		#hd[0].indreg.iFEATURES_iERROR
	.long 	0x01f2		#hd[0].indreg.iSTCONT
	.long	0x01f3		#hd[0].indreg.iSECT_N
	.long	0x01f4		#hd[0].indreg.iCYL_L_N
	.long	0x01f5		#hd[0].indreg.iCYL_H_N
	.long	0x01f6		#hd[0].indreg.iHD_N_iDRV_HD
	.long	0x03f6		#hd[0].indreg.iALT_STS_iDEV_CTRL
	.long	0		#hd[0].disco[0].{presente,dma} (sono byte!)
	.word	0		#hd[0].disco[0].geom.cil
	.word	0		#hd[0].disco[0].geom.test
	.long	0		#hd[0].disco[0].geom.sett
	.long	0		#hd[0].disco[0].geom.tot_sett
	.long	0		#hd[0].disco[1].{presente,dma} (sono byte!)
	.word	0		#hd[0].disco[1].geom.cil
	.word	0		#hd[0].disco[1].geom.test
	.long	0		#hd[0].disco[1].geom.sett
	.long	0		#hd[0].disco[1].geom.tot_sett
	.long	0		#hd[0].comando
				#hd[0].errore (byte!)
	.long	0		#hd[0].cont (byte!)
	.long	0		#hd[0].punt
	.long	0		#hd[0].mutex
	.long	0		#hd[0].sincr
	.long	0x0177		#hd[1].indreg.iCMD_iSTS
	.long	0x0170		#hd[1].indreg.iDATA
	.long	0x0171		#hd[1].indreg.iFEATURES_iERROR
	.long 	0x0172		#hd[1].indreg.iSTCONT
	.long	0x0173		#hd[1].indreg.iSECT_N
	.long	0x0174		#hd[1].indreg.iCYL_L_N
	.long	0x0175		#hd[1].indreg.iCYL_H_N
	.long	0x0176		#hd[1].indreg.iHD_N_iDRV_HD
	.long	0x0376		#hd[1].indreg.iALT_STS_iDEV_CTRL
	.long	0		#hd[1].disco[0].{presente,dma} (sono byte!)
	.word	0		#hd[1].disco[0].geom.cil
	.word	0		#hd[1].disco[0].geom.test
	.long	0		#hd[1].disco[0].geom.sett
	.long	0		#hd[1].disco[0].geom.tot_sett
	.long	0		#hd[1].disco[1].{presente,dma} (sono byte!)
	.word	0		#hd[1].disco[1].geom.cil
	.word	0		#hd[1].disco[1].geom.test
	.long	0		#hd[1].disco[1].geom.sett
	.long	0		#hd[1].disco[1].geom.tot_sett
	.long	0		#hd[1].comando
				#hd[1].errore (byte!)
	.long	0		#hd[1].cont (byte!)
	.long	0		#hd[1].punt
	.long	0		#hd[1].mutex
	.long	0		#hd[1].sincr

################################################################################
##                            SEZIONE TESTO                                   ##
################################################################################

	.text

################################################################################
##                          CHIAMATE DI SISTEMA                               ##
################################################################################

	.global _terminate_p
_terminate_p:
	int $tipo_t
	ret

	.global _sem_ini
_sem_ini:
	int $tipo_si
	ret

	.global _sem_wait
_sem_wait:
	int $tipo_w
	ret

	.global _sem_signal
_sem_signal:
	int $tipo_s
	ret

################################################################################
##                     INTERFACCIA VERSO IL MODULO SISTEMA                    ##
################################################################################

	.global _activate_pe
_activate_pe:
	int $tipo_ae
	ret

	.global _nwfi
_nwfi:
	int $tipo_nwfi
	ret

	.global _verifica_area
_verifica_area:
	int $tipo_va
	ret

	.global _panic
_panic:
	int $tipo_p
	ret

	.global _fill_gate
_fill_gate:
	int $tipo_fg
	ret

	.global _reboot
_reboot:
	int $tipo_r
	ret

	.global _con_write
_con_write:
	int $tipo_cw
	ret

	.global _con_read
_con_read:
	int $tipo_cr
	ret

	.global _con_save
_con_save:
	int $tipo_cs
	ret

	.global _con_load
_con_load:
	int $tipo_cl
	ret

	.global _con_update
_con_update:
	int $tipo_cu
	ret

	.global _con_init
_con_init:
	int $tipo_ci
	ret
 
	.global _trasforma
_trasforma:
	int $tipo_tras
	ret
	
	.global _printk
_printk:
	int $tipo_pk
	ret	
	
	.global _delay
_delay:
	int $tipo_d
	ret

################################################################################
##                         FUNZIONI DI SUPPORTO                               ##
################################################################################

# Ingresso di un byte da una porta di IO
# Replicato in sistema.s
#
	.global _inputb
_inputb:
	pushl %eax
	pushl %edx
	movl 12(%esp), %edx
	inb %dx, %al
	movl 16(%esp), %edx
	movb %al, (%edx)
	popl %edx
	popl %eax
	ret

# Uscita di un byte su una porta di IO
# Replicato in sistema.s
#
	.global _outputb
_outputb:
	pushl %eax
	pushl %edx
	movb 12(%esp), %al
	movl 16(%esp), %edx
	outb %al, %dx
	popl %edx
	popl %eax
	ret

# Ingresso di una word da una porta di IO
# Replicato in sistema.s
#
	.global _inputw
_inputw:
	pushl %eax
	pushl %edx
	movl 12(%esp), %edx
	inw %dx, %ax
	movl 16(%esp), %edx
	movw %ax, (%edx)
	popl %edx
	popl %eax
	ret

# Uscita di una word su una porta di IO
# Replicato in sistema.s
#
	.global _outputw
_outputw:
	pushl %eax
	pushl %edx
	movw 12(%esp), %ax
	movl 16(%esp), %edx
	outw %ax, %dx
	popl %edx
	popl %eax
	ret

# Ingresso di un buffer continuo da un buffer di IO di word
# Replicato in sistema.s
#
	.global _inputbuffw
_inputbuffw:
	pushl %eax
	pushl %edx
	pushl %edi
	pushl %ecx
	
	movl 20(%esp), %edx
	movl 24(%esp), %edi
	movl 28(%esp),%ecx
	cld
	rep
	insw
	
	popl %ecx
	popl %edi
	popl %edx
	popl %eax
	ret

# Uscita di un buffer continuo su una buffer di IO di word
# Replicato in sistema.s
#
	.global _outputbuffw
_outputbuffw:
	pushl %eax
	pushl %edx
	pushl %esi
	pushl %ecx
	
	movl 24(%esp), %edx
	movl 20(%esp), %esi
	movl 28(%esp),%ecx
	cld
	rep
	outsw
	
	popl %ecx
	popl %esi
	popl %edx
	popl %eax
	ret

# Indirizzi delle porte del controllore delle interruzioni
#
.set OCW1M, 0x21
.set OCW2M, 0x20
.set OCW1S, 0xa1
.set OCW2S, 0xa0
.set EOI, 0x20

# Inizio dell' ingresso da una interfaccia seriale
#
	.global _go_inputse
_go_inputse:
	pushl %eax
	pushl %edx

	movl 12(%esp), %edx		# ind. di IER in edx
	inb %dx, %al
	orb $0x01, %al			# abilitazione dell' interfaccia a
					#  generare interruzioni in ingresso
	outb %al, %dx

	cmpw $0x03f9, %dx
	jne 1f
	movb $0b11101111, %ah
	jmp 2f

1:
	movb $0b11110111, %ah
2:
	cli
	inb $OCW1M, %al
	andb %ah, %al
	outb %al, $OCW1M		# abilitazione del controllore ad
					#  accettare richieste di interruzione
					#  da parte dell' interfaccia
	sti

	popl %edx
	popl %eax
	ret

# Fine dell' ingresso da un' interfaccia seriale
#
	.global _halt_inputse
_halt_inputse:
	pushl %eax
	pushl %edx

	movl 12(%esp), %edx		# ind. di IER in edx
	inb %dx, %al
	and $0xfe, %al
	outb %al, %dx			# disabilitazione della generazione
					#  di interruzioni

	cmpw $0x03f9, %dx
	jne 1f
	movb $0b00010000, %ah
	jmp 2f

1:
	movb $0b00001000, %ah
2:
	cli
	inb $OCW1M, %al
	andb %ah, %al
	outb %al, $OCW1M		# disabilitazione dell 'interruzione al
					#  controllore
	sti

	popl %edx
	popl %eax
	ret

# Inizio dell' uscita su interfaccia seriale
#
	.global _go_outputse
_go_outputse:
	pushl %eax
	pushl %edx

	movl 12(%esp), %edx		# ind. di IER in edx
	inb %dx, %al
	orb $0x02, %al
	outb %al, %dx

	cmpw $0x03f9, %dx
	jne 1f
	movb $0b11101111, %ah
	jmp 2f

1:
	movb $0b11110111, %ah
2:
	cli
	inb $OCW1M, %al
	andb %ah, %al
	outb %al, $OCW1M
	sti

	popl %edx
	popl %eax
	ret

# Fine dell' uscita su interfaccia seriale
#
	.global _halt_outputse
_halt_outputse:
	pushl %eax
	pushl %edx

	movl 12(%esp), %edx		# ind. di IER in edx
	inb %dx, %al
	and $0xfd, %al
	outb %al, %dx

	cmpw $0x03f9, %dx
	jne 1f
	movb $0b00010000, %ah
	jmp 2f

1:
	movb $0b00001000, %ah
2:
	cli
	inb $OCW1M, %al
	andb %ah, %al
	outb %al, $OCW1M
	sti

	popl %edx
	popl %eax
	ret

# Indirizzi delle porte delle interfacce seriali
#
.set LCR1, 0x03fb
.set LCR2, 0x02fb
.set DLR_LSB1, 0x03f8
.set DLR_LSB2, 0x02f8
.set DLR_MSB1, 0x03f9
.set DLR_MSB2, 0x02f9
.set IER1, 0x03f9
.set IER2, 0x02f9
.set RBR1, 0x03f8
.set RBR2, 0x02f8
.set MCR1, 0x03fc
.set MCR2, 0x02fc


# Inizializzazione delle interfacce seriali
# 
	.global _com_setup
_com_setup:
	pushl %eax
	pushl %edx

	movb $0x80, %al
	movw $LCR1, %dx
	outb %al, %dx
	movw $0x000c, %ax
	movw $DLR_LSB1, %dx
	outb %al, %dx
	movb %ah, %al
	movw $DLR_MSB1, %dx
	outb %al, %dx
	movb $0x03, %al
	movw $LCR1, %dx
	outb %al, %dx
	movb $0x00, %al
	movw $IER1, %dx
	outb %al, %dx
	movw $RBR1, %dx
	movw $MCR1, %dx			# abilitazione porta 3-state
	movb $0b00001000, %al
	outb %al, %dx
	inb %dx, %al

	movb $0x80, %al
	movw $LCR2, %dx
	outb %al, %dx
	movw $0x000c, %ax
	movw $DLR_LSB2, %dx
	outb %al, %dx
	movb %ah, %al
	movw $DLR_MSB2, %dx
	outb %al, %dx
	movb $0x03, %al
	movw $LCR2, %dx
	outb %al, %dx
	movb $0x00, %al
	movw $IER2, %dx
	outb %al, %dx
	movw $RBR2, %dx
	movw $MCR2, %dx
	movb $0b00001000, %al
	outb %al, %dx

	inb %dx, %al

	popl %edx
	popl %eax

	ret

# Abilitazione del controllore delle interruzioni ad inoltrare le richieste
#  provenienti dalla tastera
#
	.global _abilita_tastiera
_abilita_tastiera:
	inb $OCW1M, %al
	andb $0b11111101, %al
	outb %al, $OCW1M
	ret

# Funzioni di utilita' per il meccanismo di DMA
#

.set DMA_MODE2,	0x00D6
.set DMA_CMD1,	0x0008
.set DMA_CMD2,	0x00D0
.set DMA_RMASK1, 0x000F
.set DMA_RMASK2, 0x00DE
.set DMA_REQ1, 0x0009
.set DMA_REQ2, 0x00D2

# Inizializzazione del controllore DMA eseguita all'avvio del sistema
#
	.global _dma_init
_dma_init:
	pushl %eax
	pushl %ecx
	movb $0b00001111,%al
	outb %al,$DMA_RMASK1		# maschera tutti i canali
	movb $0b00001111,%al
	outb %al,$DMA_RMASK2
	movb $0b11000000,%al		# canale 4 (0bxxxxxx00)
	outb %al,$DMA_MODE2		#  in cascade (0b11xxxxxx)
	
	movb $0b00001110,%al		# rimuove la maschera per il canale 4
	outb %al,$DMA_RMASK2
	
	movb $0,%cl
di_cy1:	cmpb $4,%cl			# Annulla i trasferimenti
	je dcy1_e
	movb $0x00,%al
	orb %cl,%al
	outb %al,$DMA_REQ1
	outb %al,$DMA_REQ2
	incb %cl
	jmp di_cy1
dcy1_e:	inb $DMA_CMD1,%al		# elimina i TerminalCount
	inb $DMA_CMD2,%al
	
#	movb $0b11011000,%al		# non si effettua alcuna programmazione
#	outb %al,$DMA_CMD1		#  e` gia` stata effettuata dal BIOS
#	outb %al,$DMA_CMD2

	popl %ecx
	popl %eax
	ret

# reset del flip flop interno al controllore DMA per accedere ai
#  registri "doppi" partendo da una situazione nota
#
.set DMA_M2, 0x00C0
.set DMA_CLEAR1, 0x000C
.set DMA_CLEAR2, 0x00D8

clear_FF:
	pushl %edx
	pushl %eax
	movl 12(%esp),%eax
	andw $DMA_M2,%ax
	cmpw $DMA_M2,%ax	# capisce se lavorare sul master o sullo slave
	je dmc_2
	jmp dmc_1
dmc_1:	movw $DMA_CLEAR1,%dx
	jmp dmc_e
dmc_2:	movw $DMA_CLEAR2,%dx		
dmc_e:	outb %al,%dx		# reset del giusto flip-flop
	popl %eax
	popl %edx
	ret

# Inizializzazione dell'indirizzo per un trasferimento in DMA, scompone
#  l'indirizzo passato e lo invia ai registri corretti nell'ordine corretto
#
	.global _setup_addr_dma
_setup_addr_dma:
	pushl %eax
	pushl %edx
	movl 20(%esp),%eax
	movl 16(%esp),%edx
	pushl %edx		# reset del flip-flop per l'accesso a iMAR
	call clear_FF
	addl $4,%esp
	outb %al,%dx		# caricato iMAR_MSB
	movb %ah,%al
	outb %al,%dx		# caricato iMAR_LSB
	shrl $16,%eax
	movl 12(%esp),%edx
	outb %al,%dx		# caricato iPAGE
	popl %edx
	popl %eax
	ret

# Inizializzazione del contatore per un trasferimento in DMA, scompone
#  il valore passato e accede i registri nell'ordine corretto
	.global _setup_cont_dma
_setup_cont_dma:
	pushl %eax
	pushl %edx
	movl 16(%esp),%eax
	decl %eax		# decrementa di uno, cosi` lavora il DMA
	movl 12(%esp),%edx
	pushl %edx		# reset del flip-flop per l'accesso a iTCR
	call clear_FF
	addl $4,%esp
	outb %al,%dx		# caricato iTCR_MSB
	movb %ah,%al
	outb %al,%dx		# caricato iTCR_LSB
	popl %edx
	popl %eax
	ret

# Avvia una operazione di trasferimento in DMA in ingresso, programmando
#  un canale specificato in lettura secondo una modalita` di funzionamento
#
	.global _go_moder_in
_go_moder_in:
	pushl %eax
	pushl %edx
	movb 12(%esp),%al
	movb 24(%esp),%ah
	andb $0x03,%ah		# selezione del canale
	addb %ah,%al
	orb $0b00000100,%al	# Maschera di lettura, 0bxxxx01xx
	movl 16(%esp),%edx
	outb %al,%dx
	movl 20(%esp),%edx
	movb 24(%esp),%al
	andb $0b11111011,%al	# Smaschera il canale
	outb %al,%dx
	outb %al,%dx		# inviata due volte(?)
	popl %edx
	popl %eax
	ret

# Avvia una operazione di trasferimento in DMA in uscita, programmando
#  un canale specificato in scrittura secondo una modalita` di funzionamento
#
	.global _go_moder_out
_go_moder_out:
	pushl %eax
	pushl %edx
	movb 12(%esp),%al
	movb 24(%esp),%ah	# selezione del canale
	andb $0x03,%ah
	addb %ah,%al
	orb $0b00001000,%al	# Maschera di scrittura, 0bxxxx10xx
	movl 16(%esp),%edx
	outb %al,%dx
	movl 20(%esp),%edx
	movb 24(%esp),%al
	andb $0b11111011,%al	# Smaschera il canale
	outb %al,%dx
	outb %al,%dx		# Ripetuto due volte(?)
	popl %edx
	popl %eax	
	ret

# Interrompe un trasferimento o, se gia` terminato, resetta il canale
#  e ne riattiva la maschera
#
	.global _halt_moder
_halt_moder:
	pushl %eax
	pushl %edx
	movb $0b00000000,%al
	movb 20(%esp),%ah
	orb %ah,%al		# Selezione di un canale
	movl 12(%esp),%edx
	outb %al,%dx		# Abort dei trasferimenti
	movl 16(%esp),%edx
	movb %ah,%al
	orb $0b00000100,%al	# Maschera il canale
	outb %al,%dx
	outb %al,%dx		# ripetuta due volte(?)
	popl %edx
	popl %eax
	ret

# Funzioni di utilita' per la gestione del floppy disk
#

# Valuta, accedendo la CMOS del sistema la presenza di uno o piu`
#  lettori di floppy disk
# 
	.global _test_fd
_test_fd:
	pushl %edx
	movw $0x70,%dx
	movb $0x10,%al
	outb %al,%dx		# Indirizzamento della CMOS
	movw $0x71,%dx
	inb %dx,%al		# lettura del registro specificato
	cmpb $0x40,%al
	je f1d
	cmpb $0x44,%al
	je f2d
	movl $0x00,%eax
	jmp fdr
f1d:	movl $1,%eax		# ritorno: 1 solo floppy
	jmp fdr
f2d:	movl $2,%eax		# ritorno: 2 floppy installati
fdr:	popl %edx
	ret
	
# abilita il controllore di interruzioni ad inoltrare richieste
#  provenienti dal floppy disk, linea 6, controllore Master
#
	.global _abilita_fd
_abilita_fd:
	inb $OCW1M, %al
	andb $0b10111111, %al
	outb %al, $OCW1M
	ret

# Accende il motorino del drive specificato nel controllore floppy
#
	.global _motore_on
_motore_on:
	pushl %edx
	pushl %eax
	pushl %ecx
	pushl %edi
	movl 24(%esp),%edi
	cmpb $0x00,(%edi)	# legge lo stato attuale del motorino (descrittore)
	jne mon_r		# motore gia` acceso
	movb $0x01,(%edi)	# motore spento, cambia lo stato in acceso
	movb 28(%esp),%ah
	movb $0b00001100,%al
	orb %ah,%al
	cmpb $1,%ah
	je mon_1
	orb $0b00010000,%al	# prepara il comando giusto in base al drive
	jmp mon_ot
mon_1:	orb $0b00100000,%al	
mon_ot:	movl 20(%esp),%edx
	outb %al,%dx		# invia il comando di accensione
	movw $0x7FFF,%cx
mon_l:	inb $0x80,%al
	decw %cx
	jnz mon_l		# attesa attiva che il motorino sia acceso
mon_r:	popl %edi
	popl %ecx
	popl %eax
	popl %edx
	ret

# Spenge il motorino del drive selezionato nel controllore floppy
#
	.global _motore_off
_motore_off:
	pushl %edx
	pushl %eax
	pushl %edi
	movl 20(%esp),%edi
	cmpb $0x00,(%edi)	# legge lo stato del motorino (dal descrittore)
	je mof_r		# motore gia` spento
	movb $0x00,(%edi)
	movl 16(%esp),%edx
	movb $0b00001100,%al	# prepara il comando (piu` semplice!)
	outb %al,%dx		# invia il comando
mof_r:	popl %edi
	popl %eax
	popl %edx
	ret

# Funzioni di utilita' per la gestione degli hard disk
#

# Esegue un probing del canale ATA specificato tentando di capire se
#  ad esso sono collegati nessuno, uno o piu` dispositivi
#
	.global _test_canale
_test_canale:
	pushl %ebx
	pushl %edx
	movw 12(%esp),%dx
	movw 0xFFFF,%bx
chcyc1:	inb %dx,%al
	cmpb $0xFF,%al
	je chfl1
	inb %dx,%al
	testb $0x80,%al
	jz chprb1
	decw %bx
	jnz chcyc1
chprb1:	movb $0x55,%al
	movw 16(%esp),%dx 
	outb %al,%dx
	movb $0xAA,%al
	movw 20(%esp),%dx
	outb %al,%dx
	movw 16(%esp),%dx
	outb %al,%dx
	movb $0x55,%al
	movw 20(%esp),%dx
	outb %al,%dx
	movw 16(%esp),%dx
	outb %al,%dx
	movb $0xAA,%al
	movw 20(%esp),%dx
	outb %al,%dx
	movw 16(%esp),%dx
	inb %dx,%al
	movb %al,%ah
	movw 20(%esp),%dx
	inb %dx,%al
	cmpw $0x55AA,%ax
	jne chfl1
chnfl1:	
	movl $1,%eax
	jmp chret1
chfl1:	
	movl $0,%eax
chret1:	popl %edx
	popl %ebx
	ret

# Assmendo che il canale sia "attivo" cerca di capire se il dispositivo
#  indirizzato e` un ATA (hard disk) o ATAPI (chrom)
#
	.global _leggi_signature
_leggi_signature:
	pushl %edx
	movw 8(%esp),%dx
	inb %dx,%al
	cmpb $0x01,%al
	jne sgerr
	movw 12(%esp),%dx
	inb %dx,%al
	cmpb $0x01,%al
	jne sgerr
	movw 16(%esp),%dx
	inb %dx,%al
	cmpb $0x00,%al
	je sgata1
	cmpb $0x14,%al
	je sgpi1
	jmp sgerr
sgata1:	movw 20(%esp),%dx
	inb %dx,%al
	cmpb $0x00,%al
	jne sgerr
	jmp sgata
sgpi1:	movw 20(%esp),%dx
	inb %dx,%al
	cmpb $0xEB,%al
	jne sgerr
	jmp sgpi
sgata:	
	movl $0,%eax
	jmp sgret
sgpi:	
	movl $1,%eax
	jmp sgret
sgerr:	
	movl $-1,%eax
	jmp sgret
sgret:	
	popl %edx
	ret

	.global _drive_reset
_drive_reset:
	pushl %edx
	movw 8(%esp),%dx
	movb $0x0E,%al
	outb %al,%dx
	movw 12(%esp),%dx
	inb %dx,%al
	testb $0x80,%al
	jz rserr
	movw 8(%esp),%dx
	movb $0x02,%al
	outb %al,%dx
	movl $1,%eax
	jmp rsret
rserr:	movl $0,%eax
rsret:	popl %edx
	ret

# Abilita il controller ATA specificato a generare richieste di interruzione
#  utili sia per l'ingresso che per l'uscita
#
	.global _go_inouthd
_go_inouthd:
	pushl %eax
	pushl %edx

	movl 12(%esp), %edx		# ind. di DEV_CTL in edx
	movb $0x08,%al
	outb %al, %dx			# abilitazione dell' interfaccia a
					# generare interruzioni
	popl %edx
	popl %eax
	ret

# Disabilita il controller ATA specificato a generare richeste di interruzione
#  NB. Su un PC (non su Bochs) questa stassa procedura produrra` una richiesta
#  di interruzione!
#
	.global _halt_inouthd
_halt_inouthd:
	pushl %eax
	pushl %edx

	movl 12(%esp), %edx		# ind. di DEV_CTL in edx
	movb $0x0A,%al
	outb %al, %dx			# disabilitazione della generazione
					# di interruzioni
	popl %edx
	popl %eax
	ret

# Indica al controllore quale drive deve essere selezionato
#
	.global _set_drive
_set_drive:
	pushl %eax
	pushl %edx

	movl 12(%esp),%eax
	cmpl $0,%eax
	je shd_ms
shd_sl:	movb $0xF0,%al
	jmp ms_out
shd_ms:	movb $0xE0,%al
ms_out:	movl 16(%esp),%edx
	outb %al,%dx

	popl %edx
	popl %eax
	ret

# Legge lo stato attuale del controllore per capire quale drive e` attivo
#
	.global _get_drive
_get_drive:
	pushl %eax
	pushl %edx

	movl 12(%esp),%edx
	inb %dx,%al
	
	testb $0x10,%al
	jz ghd_ms
ghd_sl:	movl $1,%eax
	jmp ms_ret
ghd_ms:	movl $0,%eax
ms_ret:	movl 16(%esp),%edx
	movl %eax,(%edx)

	popl %edx
	popl %eax
	ret
	
# Indirizzi dei registri del controller ata
#

.set HD_STS_1, 0x01F7
.set HD_STS_2, 0x0177
.set HD_SECT_N, 0x01f3
.set HD_CYL_L_N, 0x01f4
.set HD_CYL_H_N, 0x01f5
.set HD_N, 0x01f6

# Maschere dei bit fndamentali
#
.set HD_STS_DRDY, 0x40

# Abilita il controllore di interruzione ad inoltrere le richieste provenienti
#  da uno dei due controller ATA, Slave PIC, linee 14 o 15
#
	.global _abilita_hd
_abilita_hd:
	pushl %eax
	movl 8(%esp),%eax
	cmpl $0x01F7,%eax
	je umsk0
	cmpl $0x0177,%eax
	je umsk1		# Capisce qual'e` il canale
	jmp um_end
umsk0:	inb $OCW1S,%al
	andb $0b10111111,%al	# linea 14
	outb %al,$OCW1S	
	jmp um_end
umsk1:	inb $OCW1S,%al
	andb $0b01111111,%al	# linea 15
	outb %al,$OCW1S
um_end:	popl %eax
	ret
	
# Scompone correttamente il blocco iniziale di una operazione su hard disk
# e lo scrive nei registri opportuni
#
	.global _setup_addr_hd
_setup_addr_hd:
	pushl %eax
	pushl %edx
	pushl %edi
	
	movl 20(%esp),%eax
	movl 16(%esp),%edi	# accede al descrittore per comodita`
	movl 16(%edi),%edx
	outb %al,%dx		# caricato iSECT_N
	movl 20(%edi),%edx
	movb %ah,%al
	outb %al,%dx		# caricato cyl_LSB
	shrl $16,%eax
	movl 24(%edi),%edx
	outb %al,%dx		# caricato cyl_MSB
	movl 28(%edi),%edx
	inb %dx,%al		# iDRV_HD in %al
	andb $0xF0,%al		# maschera per l'indirizzo in DRV_HD
	andb $0x0F,%ah		# maschera per i 4 bit +sing di primo
	orb %ah,%al
	outb %al,%dx		# caricato iDRV_HD
	
	popl %edi
	popl %edx
	popl %eax
	ret


# Costanti per i parametri attuali di _fill_gate
# Devono coincidere con i valori specificati in sistema.s
#
.set IDT_DPL3, 0x6000
.set IDT_TYPE_TRAP, 0x0f00

# Chiama _fill_gate con i parametri specificati
#
.macro fill_io_gate gate off
	pushl $IDT_DPL3
	pushl $IDT_TYPE_TRAP
	pushl $\off
	pushl $\gate
	call _fill_gate
	addl $16, %esp
.endm

# Inizializzazione dei gate per le primitive di IO
#
	.global _fill_io_gates
_fill_io_gates:
	fill_io_gate io_tipo_rsen a_readse_n
	fill_io_gate io_tipo_rseln a_readse_ln
	fill_io_gate io_tipo_wsen a_writese_n
	fill_io_gate io_tipo_wse0 a_writese_0
	fill_io_gate io_tipo_tr a_term_read_n
	fill_io_gate io_tipo_tw a_term_write_n
	fill_io_gate io_tipo_geom a_geometria
	fill_io_gate io_tipo_rhdn a_readhd_n
	fill_io_gate io_tipo_whdn a_writehd_n
	fill_io_gate io_tipo_rfdn a_readfd_n
	fill_io_gate io_tipo_wfdn a_writefd_n
	fill_io_gate io_tipo_swp a_swap
	fill_io_gate io_tipo_sams a_setAllMixerSetting
	fill_io_gate io_tipo_gams a_getAllMixerSetting		
	fill_io_gate io_tipo_wsb a_writesb_n		
	fill_io_gate io_tipo_rsb a_readsb_n				
	fill_io_gate io_tipo_sms a_setMixerSetting				
	fill_io_gate io_tipo_gms a_getMixerSetting					
	fill_io_gate io_tipo_spon a_speakerOn
	fill_io_gate io_tipo_spof a_speakerOff	
	fill_io_gate io_tipo_dpl a_DirectPlay
	fill_io_gate io_tipo_drd a_DirectRecord	
	fill_io_gate io_tipo_gss a_getSpeakerStatus
	fill_io_gate io_tipo_rmd a_readMidi
	fill_io_gate io_tipo_wmd a_writeMidi	
	fill_io_gate io_tipo_rsmd a_resetMidi	
	fill_io_gate io_tipo_stmd a_startMidi	
	ret
	
################################################################################
##                              FUNZIONI SB16                                 ##
################################################################################
#
#funzioni di ulitlit per accedere alla sb16 linea x controllore master
#
	.global _abilita_sb
_abilita_sb:
	pushl %edx
	movb 8(%esp), %dl
	inb $OCW1M, %al
	andb %dl, %al
	outb %al, $OCW1M
	popl %edx
	ret

		
################################################################################
##                              PRIMITIVE DI IO                               ##
################################################################################

	.global a_readse_n
	.extern _c_readse_n
a_readse_n:
	salva_registri
	copia_param 4 7
	call _c_readse_n
	addl $16, %esp
	carica_registri
	iret

	.global a_readse_ln
	.extern _c_readse_ln
a_readse_ln:
	salva_registri
	copia_param 4 7
	call _c_readse_ln
	addl $16, %esp
	carica_registri
	iret

	.global a_writese_n
	.extern _c_writese_n
a_writese_n:
	salva_registri
	copia_param 3 7
	call _c_writese_n
	addl $12, %esp
	carica_registri
	iret

	.global a_writese_0
	.extern _c_writese_0	# non _c_writese_ln, che va lo stesso
a_writese_0:
	salva_registri
	copia_param 3 7
	call _c_writese_0
	addl $12, %esp
	carica_registri
	iret

	.global a_term_read_n
	.extern _c_term_read_n
a_term_read_n:	# routine int $io_tipo_tr
	salva_registri
	copia_param 3 7
	call _c_term_read_n
	addl $12, %esp
	carica_registri
	iret

	.global a_term_write_n
	.extern _c_term_write_n
a_term_write_n: # routine int $io_tipo_tw
	salva_registri
	copia_param 3 7
 	call _c_term_write_n
 	addl $12, %esp
 	carica_registri
	iret
	
	.global a_geometria
	.extern _c_geometria
a_geometria:	# routine int $io_tipo_geom
	salva_registri
	copia_param 7 7
	call _c_geometria
	addl $28,%esp
	carica_registri
	iret

	.global a_readfd_n
	.extern _c_readfd_n
a_readfd_n:
	salva_registri
	copia_param 5 7
	call _c_readfd_n
	addl $20, %esp
	carica_registri
	iret

	.global a_writefd_n
	.extern _c_writefd_n
a_writefd_n:
	salva_registri
	copia_param 5 7
	call _c_writefd_n
	addl $20, %esp
	carica_registri
	iret
	
	.global a_readhd_n
	.extern _c_readhd_n
a_readhd_n:
	salva_registri
	copia_param 6 7
	call _c_readhd_n
	addl $24, %esp
	carica_registri
	iret

	.global a_writehd_n
	.extern _c_writehd_n
a_writehd_n:
	salva_registri
	copia_param 6 7
	call _c_writehd_n
	addl $24, %esp
	carica_registri
	iret

	.global a_swap
	.extern _c_swap_pag
a_swap:
	salva_registri
	copia_param 6 7
	call _c_swap_pag
	addl $24, %esp
	carica_registri
	iret
	
	.global a_getAllMixerSetting
	.extern _c_getAllMixerSetting
a_getAllMixerSetting:	# routine int $io_tipo_gams
	salva_registri
	copia_param 1 7
	call _c_getAllMixerSetting
	addl $4, %esp
	carica_registri
	iret

	.global a_setAllMixerSetting
	.extern _c_setAllMixerSetting
a_setAllMixerSetting:	# routine int $io_tipo_sams
	salva_registri
	copia_param 1 7
	call _c_setAllMixerSetting
	addl $4, %esp
	carica_registri
	iret
	

	.global a_writesb_n
	.extern _c_writesb_n
a_writesb_n:	# routine int $io_tipo_wsb
	salva_registri
	copia_param 4 7
	call _c_writesb_n
	addl $16, %esp
	carica_registri
	iret

	.global a_readsb_n
	.extern _c_readsb_n
a_readsb_n:	# routine int $io_tipo_rsb
	salva_registri
	copia_param 4 7
	call _c_readsb_n
	addl $16, %esp
	carica_registri
	iret

	.global a_setMixerSetting
	.extern _c_setMixerSetting
a_setMixerSetting:	# routine int $io_tipo_sms
	salva_registri
	copia_param 2 7
	call _c_setMixerSetting
	addl $8, %esp
	carica_registri
	iret
	
	.global a_getMixerSetting
	.extern _c_getMixerSetting
a_getMixerSetting:	# routine int $io_tipo_gms
	salva_registri
	copia_param 2 7
	call _c_getMixerSetting
	addl $8, %esp
	carica_registri
	iret

	.global a_speakerOn
	.extern _c_speakerOn
a_speakerOn:	# routine int $io_tipo_spon
	salva_registri
	copia_param 0 7
	call _c_speakerOn
	carica_registri
	iret

	.global a_speakerOff
	.extern _c_speakerOff
a_speakerOff:	# routine int $io_tipo_spof
	salva_registri
	copia_param 0 7
	call _c_speakerOff
	carica_registri
	iret
		
	.global a_DirectPlay
	.extern _c_DirectPlay
a_DirectPlay:	# routine int $io_tipo_dpl
	salva_registri
	copia_param 1 7
	call _c_DirectPlay
	addl $4, %esp
	carica_registri
	iret
	
	.global a_DirectRecord
	.extern _c_DirectRecord
a_DirectRecord:	# routine int $io_tipo_drd
	salva_registri
	copia_param 1 7
	call _c_DirectRecord
	addl $4, %esp
	carica_registri
	iret

	.global a_getSpeakerStatus
	.extern _c_getSpeakerStatus
a_getSpeakerStatus:	# routine int $io_tipo_gss
	salva_registri
	copia_param 1 7
	call _c_getSpeakerStatus
	addl $4, %esp
	carica_registri
	iret

	.global a_readMidi
	.extern _c_readMidi
a_readMidi:	# routine int $io_tipo_rmd
	salva_registri
	copia_param 1 7
	call _c_readMidi
	addl $4, %esp
	carica_registri
	iret

	.global a_writeMidi
	.extern _c_writeMidi
a_writeMidi:	# routine int $io_tipo_wmd
	salva_registri
	copia_param 1 7
	call _c_writeMidi
	addl $4, %esp
	carica_registri
	iret

	.global a_resetMidi
	.extern _c_resetMidi
a_resetMidi:	# routine int $io_tipo_rsmd
	salva_registri
	copia_param 1 7
	call _c_resetMidi
	addl $4, %esp
	carica_registri
	iret

	.global a_startMidi
	.extern _c_startMidi
a_startMidi:	# routine int $io_tipo_stmd
	salva_registri
	copia_param 1 7
	call _c_startMidi
	addl $4, %esp
	carica_registri
	iret

							
